#include "DetectLineCUDA.h"
#include <opencv2/imgcodecs.hpp>
#include "cuda_runtime.h"
#include <opencv2/cudacodec.hpp>
#include <opencv2/cudaimgproc.hpp>
#include <opencv2/cudaarithm.hpp>

using namespace cv;

DetectLineCUDA::DetectLineCUDA(int cols, int rows, std::vector<int> _angles, int MinArea, float Alpha, float Beta, float HairWidth)
{
	Width = cols;
	Height = rows;

    NumberOfFilter = 8;
    this->MinArea = MinArea;
    this->Alpha = Alpha;
    this->Beta = Beta;
    this->HairWidth = HairWidth;

	SigmaX = 8.0f * (sqrt(2.0 * log(2) / CV_PI)) * HairWidth / Alpha / Beta / CV_PI;
	SigmaY = 0.8f * SigmaX;
	KernelRadius = ceil(3.0f * SigmaX);  // sigmaX > sigamY
	KernelW = 2 * KernelRadius + 1;
	KernelH = 2 * KernelRadius + 1;
	KernelX = KernelRadius;
	KernelY = KernelRadius;
	FFTH = snapTransformSize(Height + KernelH - 1);
	FFTW = snapTransformSize(Width + KernelW - 1);

	cufftPlan2d(&_fftPlanFwd, FFTH, FFTW, CUFFT_R2C);
	cufftPlan2d(&_fftPlanInv, FFTH, FFTW, CUFFT_C2R);

    this->angles.insert(this->angles.end(), _angles.begin(), _angles.end());
}


DetectLineCUDA::~DetectLineCUDA()
{
	cufftDestroy(_fftPlanInv);
	cufftDestroy(_fftPlanFwd);
}

int DetectLineCUDA::snapTransformSize(int dataSize) {
	int hiBit;
	unsigned int lowPOT, hiPOT;

	dataSize = iAlignUp(dataSize, 16);
	for (hiBit = 31; hiBit >= 0; hiBit--) {
		if (dataSize & (1U << hiBit)) {
			break;
		}
	}
	lowPOT = 1U << hiBit;

	if (lowPOT == (unsigned int)dataSize)
		return dataSize;
	hiPOT = 1U << (hiBit + 1);
	if (hiPOT <= 1024)
		return hiPOT;
	else
		return iAlignUp(dataSize, 512);
}

int DetectLineCUDA::iAlignUp(int a, int b)
{
	return (a % b != 0) ? (a - a % b + b) : a;
}

int DetectLineCUDA::iDivUp(int a, int b)
{
	return (a % b != 0) ? (a / b + 1) : (a / b);
}

void DetectLineCUDA::detect(cv::cuda::GpuMat& src, cv::cuda::GpuMat& mask) {
	float* d_PaddedData;
	float* d_Kernel;
	float* d_PaddedKernel;
	float* d_DepthResult;
	float* d_src_c_ptr;/////////////////////////////////
	uchar* d_Result;
	cufftComplex* d_DataSpectrum;
	cufftComplex* d_KernelSpectrum;
	cufftComplex* d_TempSpectrum;

	//// allocate
    cudaMalloc((float**)&d_src_c_ptr, src.cols * src.rows * sizeof(float));////////////////////////////////////
    cudaMalloc((void**)&d_Kernel, KernelH * KernelW * NumberOfFilter * sizeof(float));
    cudaMalloc((void**)&d_PaddedData, FFTH * FFTW * sizeof(float));
    cudaMalloc((void**)&d_PaddedKernel, FFTH * FFTW * sizeof(float));
    cudaMalloc((void**)&d_DepthResult, FFTH * FFTW * angles.size() * sizeof(float));
    cudaMalloc((void**)&d_Result, Height * Width * sizeof(uchar));
    cudaMalloc((void**)&d_DataSpectrum, FFTH * (FFTW / 2 + 1) * sizeof(cufftComplex));
    cudaMalloc((void**)&d_KernelSpectrum, FFTH * (FFTW / 2 + 1) * sizeof(cufftComplex));
    cudaMalloc((void**)&d_TempSpectrum, FFTH * (FFTW / 2 + 1) * sizeof(cufftComplex));

	//// init data
	float* h_kernels = _initGaborFilterCube();
    cudaMemcpy(d_Kernel, h_kernels, KernelH * KernelW * NumberOfFilter * sizeof(float), cudaMemcpyHostToDevice);

	cuda::GpuMat src_f;
	if (src.channels() == 3) {
		cuda::GpuMat src_gray;
		src.convertTo(src_gray, CV_32FC3);
		cuda::cvtColor(src_gray, src_f,cv::COLOR_BGR2GRAY);
	}
	else {
		src.convertTo(src_f, CV_32F);
	}

	cuda::GpuMat d_src_c_ptr_mat(src.rows, src.cols, CV_32FC1, d_src_c_ptr);

	cuda::GpuMat pand_mat(FFTH, FFTW, CV_32FC1, d_PaddedData);
	cuda::copyMakeBorder(src_f, d_src_c_ptr_mat, 0, FFTH - src.rows, 0, FFTW - src.cols, BORDER_WRAP);
	d_src_c_ptr_mat.copyTo(pand_mat);

	// FFT data
	cufftExecR2C(_fftPlanFwd, (cufftReal*)d_PaddedData, d_DataSpectrum);
	cudaDeviceSynchronize();

    for (int i = 0; i < angles.size(); i++) {
        int kernel_offset = angles[i] * KernelH * KernelW;
		int data_offset = i * FFTH * FFTW;

		_padKernel(d_PaddedKernel, d_Kernel + kernel_offset);

		cv::cuda::GpuMat raw_dst_mat(FFTH, FFTW, CV_32FC1, d_PaddedKernel);
		cv::Mat cpumat;
		raw_dst_mat.download(cpumat);

		// FFT kernel
        cufftExecR2C(_fftPlanFwd, (cufftReal*)d_PaddedKernel, (cufftComplex*)d_KernelSpectrum);
        cudaDeviceSynchronize();

		//// mul
		_modulateAndNormalize(d_TempSpectrum, d_DataSpectrum, d_KernelSpectrum, 1);
        cufftExecC2R(_fftPlanInv, (cufftComplex*)d_TempSpectrum, (cufftReal*)(&d_DepthResult[data_offset]));
        cudaDeviceSynchronize();


		cv::cuda::GpuMat d_DepthResult_mat(FFTH, FFTW, CV_32FC1, d_DepthResult + data_offset);
		d_DepthResult_mat.download(cpumat);
	}

	_cubeReduction(d_DepthResult, d_Result);

	cuda::GpuMat d_Result_mat(Height , Width,CV_8UC1, d_Result);
	d_Result_mat.copyTo(mask);


	delete h_kernels;
	cudaFree(d_Kernel);
	cudaFree(d_DataSpectrum);
	cudaFree(d_KernelSpectrum);
	cudaFree(d_TempSpectrum);
	cudaFree(d_PaddedKernel);
	cudaFree(d_DepthResult);
	cudaFree(d_Result);
	cudaFree(d_PaddedData);
	cudaFree(d_src_c_ptr);///////////////////////////////
	
}


float* DetectLineCUDA::_initGaborFilterCube() {
	float* output = new float[KernelW * KernelH * NumberOfFilter];
	float* output_ptr = output;
	for (int curNum = 0; curNum < NumberOfFilter; curNum++) {
		float theta = (float)CV_PI / NumberOfFilter * curNum;
		for (int y = -KernelRadius; y < KernelRadius + 1; y++) {
			for (int x = -KernelRadius; x < KernelRadius + 1; x++, output_ptr++) {
				float xx = x;
				float yy = y;
				float xp = xx * cos(theta) + yy * sin(theta);
				float yp = yy * cos(theta) - xx * sin(theta);
				*output_ptr = exp((float)(-CV_PI) * (xp * xp / SigmaX / SigmaX + yp * yp / SigmaY / SigmaY)) * cos((float)CV_2PI * Beta / HairWidth * xp + (float)CV_PI);
			}
		}
	}
	return output;
}

